# SPDX-FileCopyrightText: 2023 Pieter Hijma <pieterhijma@noreply.codeberg.org>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import math

import FreeCAD as App
import Part

from util import *

WIDTH_TRACK = 50
DEFAULT_WIDTH_BASE = 80
WIDTH_TOP = 70
WIDTH_LEG = 15
DEFAULT_HEIGHT = 100
SPACE_PINS = 25

ANGLE_SIDE = math.atan(DEFAULT_HEIGHT / (DEFAULT_WIDTH_BASE/2 - WIDTH_TOP/2))




# the angle for a side is found in the following way: We created a loft with a
# top square of 40 mm, a base square of 80 mm.  We then found the following
# angle on the surface of such a loft.
# ANGLE_SIDE = 1.37713802635057

# We also projected it to the facing plane and we then found the following
# angle in the projection, matching our computations.

# So, this means that the angle for the side of the leg is ANGLE_SIDE and that
# the piece must be tilted ANGLE_TILT.

THICKNESS_WOOD = 3.5
WIDTH_GAP = THICKNESS_WOOD + 0.2
HEIGHT_CONNECTOR = 3

HEIGHT_SLOT = WIDTH_LEG / 2

LEFT = True
RIGHT = False

def isLegPillar(obj):
    return 'Proxy' in obj.PropertiesList and type(obj.Proxy).__name__ == 'LegPillar'

def widthBase(height):
    return WIDTH_TOP + height * 2 / (DEFAULT_HEIGHT / ((DEFAULT_WIDTH_BASE - WIDTH_TOP) / 2 ))

def heightPanel(height):
    "The actual height of the panel to reach height height when tilted"
    return height / math.sin(ANGLE_SIDE)

def dx(y, left):
    if left:
        return y / math.tan(ANGLE_SIDE)
    else:
        return -(y / math.tan(ANGLE_SIDE))

def xConnector(height, yConnector, left):
    widthB = widthBase(height)
    if left:
        return -(widthB / 2) + dx(yConnector, left)
    else:
        # checked, seems right
        return widthB / 2 - THICKNESS_WOOD + dx(yConnector, left)

def computeEdgesConnectors(obj, pBO, pTO, left):
    height = heightPanel(obj.Height)
    nrConnectors = math.floor(height / SPACE_PINS - 1)
    edges = []

    # depth_connector = THICKNESS_WOOD if left else -THICKNESS_WOOD
    # checked, seems right
    depth_connector = THICKNESS_WOOD
    # since we want to keep the corners square, we can work with these two
    
    # large and small increments based on the angle
    largeI = math.sin(ANGLE_SIDE) * depth_connector
    smallI = math.cos(ANGLE_SIDE) * depth_connector

    smallGap = (WIDTH_GAP - THICKNESS_WOOD)/2 * math.sin(ANGLE_SIDE)
    smallIGap = math.cos(ANGLE_SIDE) * WIDTH_GAP
    largeIGap = math.sin(ANGLE_SIDE) * WIDTH_GAP

    pBase = pBO

    # p: point, C connector, B base, T top, O outside, I inside
    for i in range(nrConnectors):
        yBO = (i + 1) * SPACE_PINS - (smallGap if left else smallI)
        xBO = xConnector(height, yBO, left)

        xBI = xBO + largeI
        yBI = yBO - smallI if left else yBO + smallI

        xTI = xBI + smallI if left else xBI - smallI
        yTI = yBI + largeIGap if left else yBI + largeI

        xTO = xBO + smallI if left else xBO - smallI
        yTO = yBO + largeIGap if left else yBO + largeI
        
        pCBO = mkP(xBO, yBO)
        pCBI = mkP(xBI, yBI)
        pCTI = mkP(xTI, yTI)
        pCTO = mkP(xTO, yTO)

        edges.append(mkE(pBase, pCBO))
        edges.append(mkE(pCBO, pCBI))
        edges.append(mkE(pCBI, pCTI))
        edges.append(mkE(pCTI, pCTO))
        pBase = pCTO

    edges.append(mkE(pBase, pTO))
    
    if not left:
        edges.reverse()
    return edges

def computeEdgesSlot(obj, pTLO, pTRO):
    height = heightPanel(obj.Height)
    # STLL: slot top left (left side)
    # checked, seems right
    xSLL = -(WIDTH_TRACK/2 + WIDTH_GAP)
    pSTLL = mkP(xSLL, height)
    pSBLL = mkP(xSLL, height - HEIGHT_SLOT)
    xSRL = -WIDTH_TRACK/2
    pSBRL = mkP(xSRL, height - HEIGHT_SLOT)
    pSTRL = mkP(xSRL, height)

    # xSLR = WIDTH_TRACK/2
    # pSTLR = mkP(xSLR, height)
    # pSBLR = mkP(xSLR, height - HEIGHT_SLOT)
    # xSRR = WIDTH_TRACK/2 + THICKNESS_WOOD
    # pSBRR = mkP(xSRR, height - HEIGHT_SLOT)
    # pSTRR = mkP(xSRR, height)

    pSTLR = mirrorP(pSTRL)
    pSBLR = mirrorP(pSBRL)
    pSBRR = mirrorP(pSBLL)
    pSTRR = mirrorP(pSTLL)
    
    edges = [
        mkE(pTLO, pSTLL),
        mkE(pSTLL, pSBLL),
        mkE(pSBLL, pSBRL),
        mkE(pSBRL, pSTRL),
        mkE(pSTRL, pSTLR),
        mkE(pSTLR, pSBLR),
        mkE(pSBLR, pSBRR),
        mkE(pSBRR, pSTRR),
        mkE(pSTRR, pTRO)
    ]
    
    return edges

def computeEdgesLegs(obj, pBRO, pBLO):
    pBRI = mkP(pBRO.x - WIDTH_LEG, 0)
    nrHoles = round(heightPanel(obj.Height) / DEFAULT_HEIGHT) - 1
    height = DEFAULT_HEIGHT - WIDTH_LEG if nrHoles > 0 else heightPanel(obj.Height) - WIDTH_LEG
    dxTI = height / math.tan(ANGLE_SIDE)
    pTRI = mkP(pBRO.x - WIDTH_LEG - dxTI, height)
    pTLI = mirrorP(pTRI)
    pBLI = mirrorP(pBRI)

    edges = [
        mkE(pBRO, pBRI),
        mkE(pBRI, pTRI),
        mkE(pTRI, pTLI),
        mkE(pTLI, pBLI),
        mkE(pBLO, pBLI)]

    return edges

def getNrHoles(obj):
    height = heightPanel(obj.Height)
    return math.floor(height / DEFAULT_HEIGHT) - 1

def computeEdgesLevels(obj):
    height = heightPanel(obj.Height)
    nrHoles = getNrHoles(obj)

    widthB = widthBase(obj.Height)
    xB = - widthB / 2 + WIDTH_LEG

    holesLevels = []

    for i in range(nrHoles):
        yB = (i + 1) * DEFAULT_HEIGHT
        maxYT = height - WIDTH_LEG
        yT = yB + DEFAULT_HEIGHT - WIDTH_LEG
        yT = min(yT, maxYT)
        xBL = xB + dx(yB, LEFT)
        xTL = xB + dx(yT, LEFT)

        pBL = mkP(xBL, yB)
        pBR = mirrorP(pBL)
        pTL = mkP(xTL, yT)
        pTR = mirrorP(pTL)
        # checked, seems right
        pSTR = mkP(WIDTH_GAP / 2, yB)
        pSBR = mkP(WIDTH_GAP / 2, yB - HEIGHT_SLOT)
        pSBL = mirrorP(pSBR)
        pSTL = mirrorP(pSTR)

        edges = [
            mkE(pBL, pSTL),
            mkE(pSTL, pSBL),
            mkE(pSBL, pSBR),
            mkE(pSBR, pSTR),
            mkE(pSTR, pBR),
            mkE(pBR, pTR),
            mkE(pTR, pTL),
            mkE(pTL, pBL)]

        holesLevels.append(edges)
    return holesLevels

def createLegPillar():
    doc = App.ActiveDocument
    obj = doc.addObject("Part::FeaturePython", "LegPillar")
    LegPillar(obj)
    obj.ViewObject.Proxy = 0 # This is mandatory unless we code the ViewProvider too.
    return obj

class LegPillar:

    def __init__(self, obj):
        obj.Proxy = self
        obj.addProperty("App::PropertyFloat", "Height").Height = DEFAULT_HEIGHT
        obj.addExtension('Part::AttachExtensionPython')

    def computeShape(self, obj):
        if obj.Height == 0:
            return

        # checked, seems right
        compensatedTickness = THICKNESS_WOOD / math.sin(ANGLE_SIDE)
        height = heightPanel(obj.Height)

        # points
        # B/T: base / top
        # L/R: left / right
        # O/I: outside / inside
        pBLO = mkP(-widthBase(obj.Height)/2, 0)
        pTLO = mkP(-WIDTH_TOP/2, height)
        pTRO = mkP(WIDTH_TOP/2 - compensatedTickness, height)
        pBRO = mkP(widthBase(obj.Height)/2 - compensatedTickness, 0)

        # edges
        edgesConnectorsLeft = computeEdgesConnectors(obj, pBLO, pTLO, LEFT)
        edgesSlot = computeEdgesSlot(obj, pTLO, pTRO)
        edgesConnectorsRight = computeEdgesConnectors(obj, pBRO, pTRO, RIGHT)
        edgesLegs = computeEdgesLegs(obj, pBRO, pBLO)

        edgesLevels = computeEdgesLevels(obj)

        edges = []
        edges.extend(edgesConnectorsLeft)
        #edges.append(mkE(pBLO, pTLO))
        edges.extend(edgesSlot)
        edges.extend(edgesConnectorsRight)
        #edges.append(mkE(pTRO, pBRO))
        edges.extend(edgesLegs)

        # wires
        wOutline = mkW(edges)
        wHoles = list(map(mkW, edgesLevels))

        wires = [wOutline]
        wires.extend(wHoles)

        # faces
        f = Part.Face(wires)

        # solids
        # checked, is right
        e = f.extrude(App.Vector(0, 0, THICKNESS_WOOD))

        obj.Shape = e

    def onChanged(self, obj, prop):
        ''' Do something when a property has changed '''
        if prop == "Height":
            self.computeShape(obj)

    def execute(self, obj):
        ''' Do something when doing a recomputation '''
        pass
        

