# SPDX-FileCopyrightText: 2023 Pieter Hijma <pieterhijma@noreply.codeberg.org>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import math

import FreeCAD as App
import Part

import LegPillar as leg
import BracketPillar as bracket
import TopBracketPillar

LEFT = True
RIGHT = False
NR_LEGS = 4

def createPillar():
    doc = App.ActiveDocument
    obj = doc.addObject("App::DocumentObjectGroupPython", "Pillar")
    Pillar(obj)
    obj.ViewObject.Proxy = 0 # This is mandatory unless we code the ViewProvider too.
    return obj

def getLegsPillar(obj):
    return [l for l in obj.Group if leg.isLegPillar(l)]

def placeLegs(obj):
    #displacementBase = leg.widthBase(obj.Height)/2 - leg.THICKNESS_WOOD
    # checked, seems right
    displacementBase = leg.widthBase(obj.Height)/2 - (leg.THICKNESS_WOOD / math.sin(leg.ANGLE_SIDE))
    #displacementThickness = (leg.THICKNESS_WOOD / math.sin(leg.ANGLE_SIDE)) / 2 
    displacementThickness = 0
    for i, l in enumerate(getLegsPillar(obj)):
        if i == 0:
            y = -displacementBase
            x = -displacementThickness
        elif i == 1:
            x = displacementBase
            y = -displacementThickness
        elif i == 2:
            y = displacementBase
            x = displacementThickness
        elif i == 3:
            x = -displacementBase
            y = displacementThickness
        l.AttachmentOffset.Base = App.Vector(x, y, 0)

def rotateLegs(obj):
    for i, l in enumerate(getLegsPillar(obj)):
        rX = App.Rotation()
        rX.Axis = App.Vector(1, 0, 0)
        rX.Angle = leg.ANGLE_SIDE
        # print(f"rX: {rX}")
        rZ = App.Rotation(App.Vector(0, 0, 1), i * 90)
        # rZ.Axis = App.Vector(0, 0, 1)
        # rZ.Angle = i * math.pi/2
        # print(f"rZ (with degrees): {rZ}")
        # print(f"rZ.multiply(rX): {rZ.multiply(rX)}")
        l.AttachmentOffset.Rotation = rZ.multiply(rX)
        # print(f"rotL: {l.AttachmentOffset.Rotation}")


def isBracketWithLevel(obj, level):
    return ('Proxy' in obj.PropertiesList and
            type(obj.Proxy).__name__ == 'BracketPillar' and
            obj.Level == level)

def isTopBracket(obj):
    return ('Proxy' in obj.PropertiesList and
            type(obj.Proxy).__name__ == 'TopBracketPillar')
        
def getTopBrackets(obj):
    return [b for b in obj.Group if isTopBracket(b)]

def getBrackets(obj, level):
    return [b for b in obj.Group if isBracketWithLevel(b, level)]

def createTopBrackets(obj, height):
    bl = TopBracketPillar.createTopBracketPillar()
    bl.adjustRelativeLinks(obj)
    obj.addObject(bl)

    lcs0 = obj.getObject('LCS_0')
    
    bl.Support = [(lcs0, '')]
    bl.MapMode = 'ObjectXY'
    bl.MapReversed = False

    # bl.AttachmentOffset.Base = App.Vector(-leg.WIDTH_TRACK/2, leg.THICKNESS_WOOD/2, height)
    bl.AttachmentOffset.Base = App.Vector(0, -leg.WIDTH_TRACK/2, height)
    bl.AttachmentOffset.Rotation = App.Rotation(App.Vector(1, 0, 0), 90)

    br = TopBracketPillar.createTopBracketPillar()
    br.adjustRelativeLinks(obj)
    obj.addObject(br)

    br.Support = [(lcs0, '')]
    br.MapMode = 'ObjectXY'
    br.MapReversed = False

    # br.AttachmentOffset.Base = App.Vector(leg.WIDTH_TRACK/2, leg.THICKNESS_WOOD/2, height)
    # checked, seems right
    br.AttachmentOffset.Base = App.Vector(0, leg.WIDTH_TRACK/2 + leg.THICKNESS_WOOD, height)
    br.AttachmentOffset.Rotation = App.Rotation(App.Vector(1, 0, 0), 90)

def createBrackets(obj, height, level):
    bt = bracket.createBracketPillar()
    bt.adjustRelativeLinks(obj)
    obj.addObject(bt)

    lcs0 = obj.getObject('LCS_0')
    
    bt.Support = [(lcs0, '')]
    print(f"bt.Support: {bt.Support}")
    bt.MapMode = 'ObjectXY'
    bt.MapReversed = False

    # checked, seems right
    bt.AttachmentOffset.Base = App.Vector(0, leg.THICKNESS_WOOD/2, height)
    bt.AttachmentOffset.Rotation = App.Rotation(App.Vector(1, 0, 0), 90)

    bt.Level = level

    bb = bracket.createBracketPillar()
    bb.adjustRelativeLinks(obj)
    obj.addObject(bb)

    bb.Support = [(lcs0, '')]
    print(f"bb.Support: {bt.Support}")
    bb.MapMode = 'ObjectXY'
    bb.MapReversed = False
    bb.Top = False

    # checked, seems right
    bb.AttachmentOffset.Base = App.Vector(-leg.THICKNESS_WOOD/2, 0, height)
    rx = App.Rotation(App.Vector(1, 0, 0), 90)
    rz = App.Rotation(App.Vector(0, 0, 1), 90)
    bb.AttachmentOffset.Rotation = rz * rx

    bb.Level = level

def setHeightBrackets(brackets, height):
    for b in brackets:
        if b.Top:
            # checked, seems right
            b.AttachmentOffset.Base = App.Vector(0, leg.THICKNESS_WOOD/2, height)
        else:
            # checked, seems right
            b.AttachmentOffset.Base = App.Vector(-leg.THICKNESS_WOOD/2, 0, height)

def ensureBrackets(obj):
    print('ensure')

    lcs0 = obj.getObject('LCS_0')
    if lcs0:
        height = obj.Height
        nrHoles = leg.getNrHoles(obj)
        
        # brs = getTopBrackets(obj)
        # print(f"{brs}")
        # if not brs:
        #     print('creating 0')
        #     createTopBrackets(obj, height)
        # else:
        #     print('setting 0')
        #     setHeightBrackets(brs, height)

        for level in range(1, nrHoles + 1):
            brs = getBrackets(obj, level)
            heightLevel = ((nrHoles - level + 1) * leg.DEFAULT_HEIGHT) * math.sin(leg.ANGLE_SIDE)
            if not brs:
                print(f"creating {level}")
                createBrackets(obj, heightLevel, level)
            else:
                print(f"setting {level}")
                setHeightBrackets(brs, heightLevel)

class Pillar:
    def __init__(self, obj):
        obj.Proxy = self
        obj.addProperty("App::PropertyFloat", "Height").Height = leg.DEFAULT_HEIGHT

        lcs0 = obj.newObject('PartDesign::CoordinateSystem','LCS_0')
        # lcs0.Support = [(part.Origin.OriginFeatures[0],'')]
        # lcs0.MapMode = 'ObjectXY'
        lcs0.MapReversed = False
        lcs0.adjustRelativeLinks(obj)
        obj.addObject(lcs0)
        
        for i in range(NR_LEGS):
            l = leg.createLegPillar()
            l.adjustRelativeLinks(obj)
            obj.addObject(l)
            l.Support = [(lcs0, '')]
            print(f"l.Support: {l.Support}")
            l.MapMode = 'ObjectXY'
            l.MapReversed = False
        placeLegs(obj)
        rotateLegs(obj)
        ensureBrackets(obj)


    def onChanged(self, obj, prop):
        ''' Do something when a property has changed '''
        print(f"Pillar.onChanged: {prop}")
        if prop == "Height":
            for l in getLegsPillar(obj):
                l.Height = obj.Height
            placeLegs(obj)
            rotateLegs(obj)
            ensureBrackets(obj)

    def execute(self, obj):
        ''' Do something when doing a recomputation '''
        pass






        

