// SPDX-FileCopyrightText: 2023 Pieter Hijma <pieterhijma@noreply.codeberg.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#include <AccelStepper.h>

#define PIN_DIR D6
#define PIN_STEP D7

#define PIN_MS1 D10
#define PIN_MS2 D9
#define PIN_MS3 D8

#define motorInterfaceType 1


#define PIN_IR_LED_FRONT    D1
#define PIN_IR_SENSOR_FRONT D0

#define PIN_IR_LED_BACK     D3
#define PIN_IR_SENSOR_BACK  D2

#define NR_SAMPLES  50

#define NR_SAMPLES_AVG 50

#define DETECTION_FACTOR 0.2

#define DISTANCE 5000

// global variables
int questionPrinted = 0;

// global variables motor
AccelStepper stepper = AccelStepper(motorInterfaceType, PIN_STEP, PIN_DIR);
int position = 2400;
int microstepFactor = 1;
int speed = 800;
int acceleration = 800;
int direction = HIGH;

float delayFactor = 2.0;


// for sensor front
int sumAvgFront = 0;

// for sensor back
int sumAvgBack = 0;


int get_diff_ambient(int pin_ir_led, int pin_ir_sensor) {
	int valLED = 0;
	int valAmbient = 0;

	for (int i = 0; i < NR_SAMPLES; i++) {
		digitalWrite(pin_ir_led, HIGH);
		delay(1);
		valLED += analogRead(pin_ir_sensor);
		
		digitalWrite(pin_ir_led, LOW);
		delay(1);
		valAmbient += analogRead(pin_ir_sensor);
	}
	
	return (valAmbient - valLED) / NR_SAMPLES;
}


void sample_ambient() {
	for (int i = 0; i < NR_SAMPLES_AVG; i++) {
		sumAvgFront += get_diff_ambient(PIN_IR_LED_FRONT, PIN_IR_SENSOR_FRONT);
		sumAvgBack += get_diff_ambient(PIN_IR_LED_BACK, PIN_IR_SENSOR_BACK);
	}
}


int detect_car(int pin_ir_led, int pin_ir_sensor, int *sumAvg, int test) {
	int diff = get_diff_ambient(pin_ir_led, pin_ir_sensor);
		
	float previousAvg = *sumAvg / (float) NR_SAMPLES_AVG;
	if (test) {
		Serial.print("avg: ");
		Serial.println(previousAvg);
		Serial.print("diff: ");
		Serial.println(diff);
		Serial.println("");
	}
	if (diff > previousAvg * (1 + DETECTION_FACTOR) ||
			diff < previousAvg * (1 - DETECTION_FACTOR)) {
		return 1;
	}
	else {
		*sumAvg -= previousAvg;
		*sumAvg += diff;
		return 0;
	}
}

void test_sensors() {
	Serial.println("Front:");
	while (!detect_car(PIN_IR_LED_FRONT, PIN_IR_SENSOR_FRONT, &sumAvgFront, 1));
	int timestamp_front = millis();

	Serial.println("Back:");
	while (!detect_car(PIN_IR_LED_BACK, PIN_IR_SENSOR_BACK, &sumAvgBack, 1));
	int timestamp_back = millis();

	int time = timestamp_back - timestamp_front;

	Serial.print("time: ");
	Serial.println(time);
}


void setup_car_detector(int pin_ir_led, int pin_ir_sensor) {
   pinMode(pin_ir_led, OUTPUT);
	 pinMode(pin_ir_sensor, INPUT);

}



void setup_motor() {
  pinMode(PIN_DIR, OUTPUT);
  pinMode(PIN_STEP, OUTPUT);
  digitalWrite(PIN_DIR, direction);

  pinMode(PIN_MS1, OUTPUT);
  pinMode(PIN_MS2, OUTPUT);
  pinMode(PIN_MS3, OUTPUT);
	
	microstepFactor = 4;
  stepper.setMaxSpeed(800 * microstepFactor);
  stepper.setAcceleration(800 * microstepFactor);

	position = 2400;

	setMicroStepFactor(microstepFactor);
}


void setup() {
	Serial.begin(9600);
	delay(3000);

	Serial.println("\nInitializing car detectors");
	setup_car_detector(PIN_IR_LED_FRONT, PIN_IR_SENSOR_FRONT);
	setup_car_detector(PIN_IR_LED_BACK, PIN_IR_SENSOR_BACK);
	Serial.println("sampling");
	sample_ambient();

	Serial.println("Setting up motor");
	setup_motor();

	Serial.println("\nSetup done");
}


void setMicroStepFactor(int factor) {
	microstepFactor = factor;
	switch (factor) {
	case 1:
		digitalWrite(PIN_MS1, LOW);
		digitalWrite(PIN_MS2, LOW);
		digitalWrite(PIN_MS3, LOW);
		break;
	case 2:
		digitalWrite(PIN_MS1, HIGH);
		digitalWrite(PIN_MS2, LOW);
		digitalWrite(PIN_MS3, LOW);
		break;
	case 4:
		digitalWrite(PIN_MS1, LOW);
		digitalWrite(PIN_MS2, HIGH);
		digitalWrite(PIN_MS3, LOW);
		break;
	case 8:
		digitalWrite(PIN_MS1, HIGH);
		digitalWrite(PIN_MS2, HIGH);
		digitalWrite(PIN_MS3, LOW);
		break;
	case 16:
		digitalWrite(PIN_MS1, HIGH);
		digitalWrite(PIN_MS2, HIGH);
		digitalWrite(PIN_MS3, HIGH);
		break;
	default:
		digitalWrite(PIN_MS1, LOW);
		digitalWrite(PIN_MS2, LOW);
		digitalWrite(PIN_MS3, LOW);
		microstepFactor = 1;
		break;
	}
}



void runMotor() {
	// Set the target position:
  stepper.move(position * microstepFactor);
  // Run to target position with set speed and acceleration/deceleration:
  stepper.runToPosition();

  delay(1000);
}




void lift_car() {
	while (!detect_car(PIN_IR_LED_FRONT, PIN_IR_SENSOR_FRONT, &sumAvgFront, 0));
	int timestamp_front = millis();

	while (!detect_car(PIN_IR_LED_BACK, PIN_IR_SENSOR_BACK, &sumAvgBack, 0));
	int timestamp_back = millis();

	int time = timestamp_back - timestamp_front;
	/* Serial.print("time: "); */
	/* Serial.println(time); */

	/* float speed = 0.1 / (time / 1000.0); */
	/* Serial.print("speed: "); */
	/* Serial.print(speed); */
	/* Serial.println(" m/s"); */

	//delay(time * delayFactor);

	stepper.move(-DISTANCE * microstepFactor);
	stepper.runToPosition();

	delay(3000);

	stepper.move(DISTANCE * microstepFactor);
	stepper.runToPosition();
}


void loop() {
	if (!questionPrinted) {
		Serial.println("Command: ");
		questionPrinted = 1;
	}
	if (Serial.available()) {
    String command = Serial.readStringUntil('\n');
    command.trim();

		String argString = command.substring(2);
		int arg = atoi(argString.c_str());
		
		if (command[0] == 'm') {
			stepper.setMaxSpeed(arg * microstepFactor);
			stepper.setAcceleration(acceleration * microstepFactor);
			Serial.print("speed: ");
			Serial.println(arg);
			runMotor();
		}
		else if (command[0] == 'p') {
			position = arg;
			Serial.print("position: ");
			Serial.println(position);
		}
		else if (command[0] == 'a') {
			acceleration = arg;
			Serial.print("acceleration: ");
			Serial.println(acceleration);
		}
		else if (command[0] == 'f') {
			int factor = arg;
			setMicroStepFactor(factor);
			Serial.print("microstepFactor: ");
			Serial.println(factor);
		}
		else if (command[0] == 'd') {
			delayFactor = arg / 10.0;
			Serial.print("delay factor: " );
			Serial.println(delayFactor);
		}
		else if (command[0] == 'c') {
			lift_car();
		}
		else if (command[0] == 't') {
			test_sensors();
		}

		questionPrinted = 0;
	}
}


